
from unittest import TestCase

from mapping_shortcuts.meta import create_collection_meta


class MetaTestCase(TestCase):
    def test_ok(self):
        MetaClass, collection = create_collection_meta()

        class A(metaclass=MetaClass):
            ...

        class B(metaclass=MetaClass):
            ...

        self.assertEqual(len(collection), 2)

    def test_base(self):
        class SubClass(type):
            ...

        MetaClass, collection = create_collection_meta(base=SubClass)

        class A(metaclass=MetaClass):
            ...

        class B(metaclass=MetaClass):
            ...

        self.assertEqual(len(collection), 2)
        self.assertIsInstance(A, SubClass)
        self.assertIsInstance(B, SubClass)

    def test_getter(self):

        MetaClass, collection = create_collection_meta(
            getter=lambda cls: cls.__name__,
        )

        class A(metaclass=MetaClass):
            ...

        class B(metaclass=MetaClass):
            ...

        self.assertEqual(len(collection), 2)
        self.assertIn('A', collection)
        self.assertEqual(collection['A'], A)
        self.assertIn('B', collection)
        self.assertEqual(collection['B'], B)

    def test_duplicate_ok(self):

        non_unique_key = 'NotUniqueKey'

        MetaClass, collection = create_collection_meta(
            getter=lambda cls: non_unique_key,
            raise_on_duplicate=False,
        )

        class A(metaclass=MetaClass):
            ...

        class B(metaclass=MetaClass):
            ...

        self.assertEqual(len(collection), 1)
        self.assertIn(non_unique_key, collection)
        self.assertEqual(collection[non_unique_key], B)

    def test_duplicate_exception(self):

        non_unique_key = 'NotUniqueKey'

        MetaClass, collection = create_collection_meta(
            getter=lambda cls: non_unique_key,
            raise_on_duplicate=True,
        )

        class A(metaclass=MetaClass):
            ...

        try:
            class B(metaclass=MetaClass):
                ...
        except ValueError as ex:
            self.assertTrue(str(ex).startswith('Duplication for key'))
        else:
            self.assertTrue(False, 'expected exception here')

        self.assertEqual(len(collection), 1)
        self.assertIn(non_unique_key, collection)
        self.assertEqual(collection[non_unique_key], A)




from typing import Any, Callable, Type, TypeVar

KT = TypeVar('KT')
T = TypeVar('T')
MetaCls = Type[T]
KeyGetter = Callable[[MetaCls[T]], KT]


def _meta_key_getter(obj: T) -> KT:
    return repr(obj)  # type: ignore


def create_collection_meta(
    base: MetaCls[T] = type,  # type: ignore
    getter: KeyGetter[T, KT] = _meta_key_getter,  # type: ignore
    raise_on_duplicate: bool = True
) -> tuple[MetaCls[T], dict[KT, MetaCls[T]]]:
    collection = {}  # type: dict[KT, MetaCls[T]]

    class CollectionMeta(base):  # type: ignore
        def __new__(mcs, name: str, bases: tuple[MetaCls[T], ...], attrs: dict[str, Any]) -> MetaCls[T]:
            cls = super().__new__(mcs, name, bases, attrs)
            key = getter(cls)
            if key in collection and raise_on_duplicate:
                raise ValueError(f'Duplication for key {key}')
            collection[key] = cls
            return cls

    return CollectionMeta, collection



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='mapping_shortcuts',
    version='0.9.9',
    author='Komissarov Andrey',
    author_email='Komissar.off.andrey@gmail.com',
    description='Useful shortcuts for create mappings',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/moff4/shortcuts',
    install_requires=[
    ],
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
    ],
)





import logging
import os
import pathlib
from typing import Iterable
import importlib


def get_submodules(
    path: str | pathlib.Path
) -> Iterable[str]:
    if isinstance(path, str):
        path = pathlib.Path(path)

    filenames = os.listdir(path)  # type: list[str]
    for filename in filenames:
        file = path.joinpath(filename)
        if file.is_dir():
            yield from get_submodules(file)
        if file.is_file():
            if not any(file.name.startswith(x) for x in '._') and file.name.endswith('.py'):
                yield str(file)[:-3].replace('/', '.')  # 3 == len('.py')


def load_package(
    path: str,
    logger: logging.Logger | None = None,
    raise_on_error: bool = False,
) -> None:
    """
        Load python package in RAM
    """
    logger = logger or logging.getLogger(__name__)
    for module_name in get_submodules(path.replace('.', '/')):
        try:
            print(module_name)
            importlib.import_module(module_name)
            # __import__(module_name)  # to import all classes
        except Exception as ex:
            if raise_on_error:
                raise ex from ex
            logger.exception('importing package "%s"', ex)




from unittest import TestCase

from mapping_shortcuts.decors import create_collector


class DecorTestCase(TestCase):

    def test_ok(self):
        decor, collection = create_collector()

        @decor('1')
        def f1():
            ...

        @decor('2')
        def f2():
            ...

        self.assertEqual(len(collection), 2)
        self.assertIn('1', collection)
        self.assertIn('2', collection)
        self.assertEqual(collection['1'], f1)
        self.assertEqual(collection['2'], f2)

    def test_duplicate_raise(self):
        decor, collection = create_collector()

        @decor('1')
        def f1():
            ...

        try:
            @decor('1')
            def f2():
                ...
        except ValueError as ex:
            self.assertTrue(str(ex).startswith('Duplication for key'))
        else:
            self.assertTrue(False, 'expected exception here')

        self.assertEqual(len(collection), 1)
        self.assertIn('1', collection)
        self.assertEqual(collection['1'], f1)

    def test_duplicate_ok(self):
        decor, collection = create_collector(
            raise_on_duplicate=False,
        )

        @decor('1')
        def f1():
            ...

        @decor('1')
        def f2():
            ...

        self.assertEqual(len(collection), 1)
        self.assertIn('1', collection)
        self.assertEqual(collection['1'], f2)




from typing import TypeVar, Callable

KT = TypeVar('KT')
VT = TypeVar('VT')
WrapType = Callable[[VT], VT]
DecoratorType = Callable[[KT], WrapType[VT]]
CollectionType = dict[KT, VT]


def create_collector(
    raise_on_duplicate: bool = True
) -> tuple[DecoratorType[KT, VT], CollectionType[KT, VT]]:
    def decor(key: KT) -> WrapType[VT]:
        def wrap(func: VT) -> VT:
            if key in collection and raise_on_duplicate:
                raise ValueError(f'Duplication for key {key}')
            collection[key] = func
            return func
        return wrap
    collection = {}  # type: CollectionType[KT, VT]
    return decor, collection
